// --- [ interrupt.S ] --------------------------------------------------------
//       01.02.12
//       tab = 3
// ASM Interrupt-Routinen fr die LED-Ansteuerung

#include  "avr/io.h"

// ----------------------------------------------------------------------------

.global TIMER1_COMPA_vect
.global TIMER1_CAPT_vect

// ----------------------------------------------------------------------------
// Wird bei einem CompareA aufgerufen und muss abgeschlossen sein,
// bevor sich der nchste CompareA Interrupt theoretisch ereignen kann.
// die Dauer ist: 6 + 4 + 3 * 3 + 4 + 6 = 29 Takte rcall/reti(2 * 4 ?)
// Zwischen 2 direkt aufeinanderfolgenden Compare-Werten sollten 40 CPU-Takte liegen,
// sonst findet kein Compare statt und die LED steuert voll durch, weil das Abschalten
// verpasst wurde.
// Durch ein Prescale von 8 und Compare-Werten, die mit dem Faktor 5 multipliziert werden,
// sollten 8 * 5 = 40 Takte je Compare-Wert eingehalten werden

.func TIMER1_COMPA_vect
TIMER1_COMPA_vect:
   push  R31
   push  R30
   push  R16

   lds   R31, ptr_hi                      // die Adresse des aktuellen pwm_buffer[] nach Z
   lds   R30, ptr_lo

   ld    R16, Z+                          // den nchsten Werte nach R16 laden
   sts   OCR1AH, R16                      // und in das Register schreiben: OCR1AH

   ld    R16, Z+
   sts   OCR1AL, R16                      // und in das Register schreiben: OCR1AL

   ld    R16, Z+
   out   _SFR_IO_ADDR(PORTD), R16         // den Portstatus schreiben

   sts   ptr_lo, R30                      // die Adresse des aktuellen pwm_buffer sichern
   sts   ptr_hi, R31

   pop   R16
   pop   R30
   pop   R31

   reti
.endfunc


// ---------------------------------------------------------------------------
// Wird am Ende eines jeden Taktes aufgerufen
// entscheidet, aus welcher Hlfte des pwm_buffer die Daten geholt werden,
// schreibt die ersten Werte fr OCR1A und Portstatus
// und sichert die Adresse des aktiven pwm_buffer

.func TIMER1_CAPT_vect
TIMER1_CAPT_vect:
   push  R31                              // benutzte Register sichern
   push  R30
   push  R16

   in    R16, _SFR_IO_ADDR(SREG)          // SREG sichern
   push  R16

                                          // den Timer stoppen
   ldi   R16, (1<<WGM12 | 1<<WGM13 | 0<<CS12 | 0<<CS11 | 0<<CS10 )
   sts   TCCR1B, R16

   lds   R16, pwm_update                  // if (pwm_update)
   tst   R16
   breq  TIMER1_b
                                          // (pwm_update > 0)
   clr   R16                              // pwm_update lschen
   sts   pwm_update, R16                  // 0 nach pwm_update schreiben

   in    R16, _SFR_IO_ADDR(GPIOR0)        // wechselt zwischen 0 und 1 und legt fest
   tst   R16
   brne  TIMER1_a                         // wenn nicht 0, dann springen
                                          // wenn 0, dann gehts hier weiter
   inc   R16                              // R16 ist 0, incrementieren zu 1
   out   _SFR_IO_ADDR(GPIOR0), R16        // und als GPIOR0 wieder speichern

   ldi   R30, lo8(pwm_buffer)
   ldi   R31, hi8(pwm_buffer)
   adiw  R30, 32                          // die obere Hlfte des Arrays aktivieren

   rjmp  TIMER1_x

TIMER1_a:                                 // GPIOR0 > 0
   clr   R16                              // R16  ist 1, lschen und
   out   _SFR_IO_ADDR(GPIOR0), R16        // und in GPIOR0 wieder speichern

   ldi   R30, lo8(pwm_buffer)             // die untere Hlfte des Arrays aktivieren
   ldi   R31, hi8(pwm_buffer)

   rjmp  TIMER1_x

TIMER1_b:                                 // (pwm_update == 0)

   in    R16, _SFR_IO_ADDR(GPIOR0)        // wechselt zwischen 0 und 1 und legt fest
   tst   R16
   breq  TIMER1_c                         // wenn 0, dann springen
                                          // GPIOR0 > 0
   ldi   R30, lo8(pwm_buffer)
   ldi   R31, hi8(pwm_buffer)
   adiw  R30, 32                          // die obere Hlfte des Arrays aktivieren
   rjmp  TIMER1_x

TIMER1_c:                                 // GPIOR0 = 0
   ldi   R30, lo8(pwm_buffer)             // die untere Hlfte des Arrays aktivieren
   ldi   R31, hi8(pwm_buffer)

TIMER1_x:                                 // den ersten Wert schreiben
   ld    R16, Z+                          // den nchsten Werte nach R16 laden
   sts   OCR1AH, R16                      // und in das Register schreiben: OCR1AH

   ld    R16, Z+
   sts   OCR1AL, R16

   ld    R16, Z+
   out   _SFR_IO_ADDR(PORTD), R16         // den Portstatus nach PORTD

   sts   ptr_lo, R30                      // die Adresse des aktuellen pwm_buffer sichern
   sts   ptr_hi, R31

   lds   R16, pwm_counter                 // pwm_counter laden und auf 0 prfen
   tst   R16                              // wenn er grer als 0 ist
   breq  TIMER1_x1
   dec   R16                              // dann decrementieren
   sts   pwm_counter, R16                 // und den genderten Wert speichern

TIMER1_x1:
   ldi   R16, (1<<WGM12 | 1<<WGM13 | 0<<CS12 | 1<<CS11 | 0<<CS10 )
   sts   TCCR1B, R16                      // den Timer1 wieder starten

   pop   R16
   out   _SFR_IO_ADDR(SREG), R16          // SREG wiederherstellen
   pop   R16                              // Register wiederherstellen
   pop   R30
   pop   R31

   reti
.endfunc

// --- [ eof ] ----------------------------------------------------------------

