// --- [ led_pwm12bit.c ] -----------------------------------------------------
//
// tab = 3 - berarbeitung vom 06.02.2012
//
// Ansteuerung von bis zu 8 LED's mit 12.5Bit-PWM bei 81 Hz.
// Die gewnschte Helligkeit wird mittels eines 1-Byte-Wertes pro LED bergeben.
// Die Helligkeit wird den Werten logarithmisch zugeordnet.
// Dazu ist eine Tabelle angelegt, in der 256 Wert aus dem Bereich 0 .. 6150 eingetragen sind.
// Das Programm sucht sich in der Tabelle die bentigten Werte raus und sortiert
// die Schaltzeitpunkte in der geforderten zeitlicher Reihenfolge.
// Eine Interrupt-Routine schreibt dann den jeweils nchsten Comparewert in s CompareA Register
// sowie den bereits vorbereiteten Bit-Status auf den LED-Port.
//
// Neue Helligkeitswerte knnen via TWI oder RS232bergeben werden,
// sie werden dann einen PWM-Takte spter bernommen.
// Alternativ kann ein bergang vom aktuellen LED-Status zu einem
// neu angegebenen veranlasst werden - die Dauer des bergangs bzw. die
// Anzahl der Zwischenschritte ist dabei als 1. Parameter anzugeben.
// Zustzlich kann noch ein 'Verzgerungsfaktor' eingebaut werden:
// normalerweise wird nach jedem PWM-Takt ein neuer Helligkeitswert berechnet,
// bei 81Hz dauert ein bergang (Parameter sei 255) 255/81 also ca. 3 Sekunden
// Wenn pwm_wait = n, dann wird erst nach n-Takten eine neue Helligkeit gerechnet,
// ein Helligkeitswechsel kann sich dann maximal 255/81 Sekunden * 255 = 14 Minuten hinziehen ...
// In der Zwischenzeit kann der TWI-Slave abgefragt werden:
// - liefert ein TRUE ( 1 ), dann ist das Gert beschftigt.
// Ein neuer Datensatz kann zwar angenommen werden, er wird aber erst bearbeitet, wenn derr
// der vorausgegangene Job abgeschlossen ist.
// Nur wenn der TWI-Slave ein FALSE (0) liefert, ist er bereit.
//
// ----------------------------------------------------------------------------
#define EXTERN

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "TWI_led_slave.h"
#include "global.h"
#include "rs232rx.h"

#include "ocra_tab.h"							// die PWM-Wert als PROGMEM ocra_tab[256]

uint8_t  led_value[LEDs+2];					// Array mit empfangenen 1-Byte-Wert fr die Helligkeit der LEDs
uint8_t  led_bitcode[LEDs+2];					// das zu schaltende Port-Bit fr den zugeh. Kanal
uint8_t pwm_wait = 0;                 		// verlangsamt den bergang durch Wartephasen
volatile uint8_t pwm_update;					// zeigt an, ob genderte PWM-Werte vorliegen
volatile uint8_t pwm_counter;					// zhlt PWM-Takte fr Verzgerung
uint8_t  ausgangswert[LEDs];					// die unsortierten Helligkeitswerte der letzten Einstellung,
														// bentigt fr das berblenden auf einen neuen Wert
uint8_t pwm_buffer[64];							// hier werden die vorberechneten Werte fr OCR1A und Portpins abgelegt
uint8_t portbit[8];								// Portbits zum Schalten der einzelnen LEDs

uint8_t ptr_hi, ptr_lo;                   // Variablen, die von der ISR genutzt werden, um einen
                                          // Pointer auf die folgenden OCR1A-Werte zwischenzuspeichern
char cv[10];                              // zum debuggen
// ----------------------------------------------------------------------------

void fade(uint8_t schrittzahl); 				// Berechnet die Zwischenschritte beim fading
void set_light(void);							// Bereitet die Helligkeitswerte fr die PWM-Ausgabe auf
void demo1(void);
void demo2(void);
void demo3(void);
void demo4(void);
// ----------------------------------------------------------------------------
int main(void)
{
uint8_t i, modus;

// --- Ausgnge fr die LEDs setzen in Ausgangsposition
#ifdef HIGHSIDE
LED_PORT = LED_PINS;
#else
LED_PORT = 0;
#endif

LED_DDR  = LED_PINS;

// GPIOR0 legt fest, in welcher Hlfte des Array pwm_buffer[] die vorberechneten Daten abgelegt werden.
GPIOR0    = 0;                            // die obere Hlfte benutzen

// Index Parameter = LEDPin               // Position im Kommando und zugehrige LED:
portbit[0] = LED7;								// 1. Parameter definiert LED7
portbit[1] = LED6;								// 2. Parameter definiert LED6
portbit[2] = LED5;                        // ...
portbit[3] = LED4;                        // ...
portbit[4] = LED3;                        // ...
portbit[5] = LED2;                        // ...
portbit[6] = LED1;                        // 7. Parameter definiert fr LED 1
portbit[7] = 0;									// fr RS232-Eingabe muss der Pin freibleiben

TWI_init();											// TWI initialisieren
rs232_init();										// 9600 Baud

														// Mode 12 CTC, Prescale 8
TCCR1B = ( 1<<WGM12 | 1<<WGM13 | 0<<CS12 | 1<<CS11 | 0<<CS10 );
ICR1   = PWM_MAX;									// oberes Ende des Zhlers
TIMSK1 = (1<<ICIE1 | 1<< OCIE1A );			// Interrupts freigeben
sei();

// --- Als Ausgangsposition die LED-Helligkeitswerte auf gleichen Wert setzen
for (i = 0; i < LEDs; i++)
	{
	led_value[i] = 200;                    //
	}
set_light();
fading_activ = FALSE;

// --- Hauptprogrammschleife --------------------------------------------------
while(1)
	{
	while (!(i = (TWI_ready())));				// warten bis neue Daten angekommen sind

	if (i == 2)										// das erste Byte dient als Steuerzeichen:
		{                                   // 0 -> festen Helligkeitswert setzen
														// 1 -> Variable pwm_wait verndern
														// x -> Schrittzahl fr berblendung
      if (twi_buffer[0] == 1)					// der Variablen 'pwm_wait' einen Wert zuweisen
         {
         modus = twi_buffer[1];
         TWI_clear_received();			   // TWI-Buffer und Index lschen zur Aufnahme neuer Daten

         if (modus < 251)                 // Werte < 254 werden als Wartezeit bernommen
            {
            pwm_wait = modus;		         // den Wert aus twi_buffer[1] kopieren
            }

         else if (modus == 251) demo1();   // hier werden Demos aufgerufen
         else if (modus == 252) demo2();
         else if (modus == 253) demo3();
         else if (modus == 254) demo4();
         }

		continue;                           // zurck an den Anfang von while(1)
		}

	for (i = 1; i < (LEDs + 1); i++)			// die Daten ohne Byte 0 umkopieren nach led_value[]
		{
		led_value[i - 1] = twi_buffer[i];	// die LED-Helligkeits-Wert umkopieren
		}												// ohne das Steuerbyte in led_value[0].
                                          // Das wird gleich in parameter0 umkopiert
   modus = twi_buffer[0];                 // der Wert wird noch bentigt
   TWI_clear_received();                  // den Empfangspuffer lschen und fr neue Daten bereitstellen

														// Byte 0 als Steuerbyte auswerten
	if (modus == 0)						      // wenn der Parameter == 0, dann
		{
		set_light();								// die Werte umsortieren und direkt ausgeben
		}

	else												//   i ist weder '0' noch '1' -> bergang gewnscht
		{
		fade(modus);						      // die Schrittzahl wurde in parameter0 bergeben
		}
	}
}

// ----------------------------------------------------------------------------
// Blendet von Ausgangswert (dem zuletzt benutzten Wertepaar) nach Zielwert in Schrittzahl ber.
// Verwendet die globalen Variablen:
// - ausgangswert[] (wird in  set_light() gespeichert),
// - led_value[],
// - pwm_wait.
// Zur Verbesserung der Genauigkeit/Verringerung von Rundungsfehlern
// werden die Zwischenwerte mit 128 multipliziert.
void fade(uint8_t schrittzahl)
{
uint8_t zielwert[LEDs];                 	// den Zielwert der Helligkeit speichern
int16_t zwischenwert[LEDs];               // Wert * 128 fr Zwischenschritt
int16_t delta[LEDs];                      // Differenz * 128 zwischen 2 berblend-Schritten
uint8_t i;

fading_activ = schrittzahl;

for (i = 0; i < LEDs; i++) zielwert[i] = led_value[i];
// den Abstand zwischen Ausgangs- und Zielwert ermitteln
// und mit 128 multiplizieren und durch die Schrittzahl dividieren
for (i = 0; i < LEDs; i++)
	{
	delta[i] = led_value[i] - ausgangswert[i]; // der Ausgangswert wurde beim letzten sort_byte() gespeichert
	delta[i] *= 128;
	delta[i] /= schrittzahl;
	zwischenwert[i] = (uint16_t) (ausgangswert[i] * 128);
	}

while(--schrittzahl)                      // n - 1 Schritte, weil am Ende eine Direktzuweisung des Zielwertes erfolgt
	{
	fading_activ = schrittzahl;
	for (i = 0; i < LEDs; i++)  				// Nchsten Zwischenwert ermitteln
		{
		zwischenwert[i] += delta[i]; 			// Zum letzten Zwischenwert die Differenz der Schritte addieren
		if (zwischenwert[i] < 0) zwischenwert[i] = 0; // Falls durch Rundungsprobleme ein Wert < 0 entsteht
		led_value[i] = (uint8_t) (zwischenwert[i] / 128); // und fr Ausgabe auf 1 Byte reduzieren

		}

	set_light();									// hier wirden die neuen OCR1A-Werte vorbereitet und zum Beginn eines
														// neuen PWM-Taktes bernommen.
	pwm_counter = pwm_wait;						// Verzgerung des fadening (pwm_wait wurde umgerechnet auf 0 - pwm_wait !)
                                          // jede ICR1-ISR decrementiert um 1
	while(pwm_counter);							// wartet, bis der Counter vom PWM-Start auf 0 decrementiert ist
	}
                                          // als letzten Wert den exakten Zielwert bergeben
														// dadurch werden Rundungsfehler ausgeglichen
for (i = 0; i < LEDs; i++) led_value[i] = zielwert[i];
set_light();
fading_activ = FALSE;
}

// ----------------------------------------------------------------------------
// den Buffer in aufsteigender Folge sortieren
// dabei auch die jeweils zugehrigen Bitcodes mitsortieren
// so dass Helligkeitswert und Bitcode den selben Indexwert behalten
// anschlieend gleichzeitige Schaltvorgnge zusammenlegen und
// die Bitmasken fr den jeweiligen Schaltzeitpunkt zusammenfhren
// benutzt die globalen Variablen
// - led_value[LEDs];	// bermittelter 1-Byte-Wert fr Helligkeit
// - led_bitcode[LEDs];	// das zu schaltende Port-Bit fr den zugeh. Kanal

void set_light(void)
{
uint8_t ready = FALSE;
uint8_t i, j, k, t, cnt;
uint16_t tt;

// Fr den Fall, dass spter ein berblenden gefordert wird,
// muss der letzte Helligkeitswert als Ausgangswert bekannt sein:
// hier wird eine Kopie angelegt
for (i = 0; i < LEDs; i++)	ausgangswert[i] = led_value[i];

// die Bits fr die Portpins der jeweiligen LEDs setzen
// led_bitcode[i] enthlt genau dasjenige Bit, das die LED[i] schaltet
for (i = 0; i < LEDs; i++) led_bitcode[i] = portbit[i];
// zustzlich noch die beiden folgenden Byte vorbelegen
led_bitcode[LEDs] = led_bitcode[LEDs + 1] = 0;
led_value[LEDs] = led_value[LEDs+1] = 255;// fr nachfolgendes Element die (n + 1).LED

cnt = (LEDs + 1);  // es ist noch eine 'virtuelle' LED fr volle Helligkeit angehngt worden, daher LED's + 1

// die LEDs und der zugehrigen led_bitcode[] in der Reihenfolge der Abschaltzeitpunkt sortieren
// das heit nach dem Helligkeitswert, kleinere Werte stehen am Anfang, weil diese zuerst abgeschaltet werden.
// Dabei auch die zugehrigen Bitcodes mitsortieren, so dass Helligkeitswert und Bitcode den selben Indexwert behalten.
// Die Sortierung erfolgt durch Vertauschen:
// immer wenn das n + 1. Element kleiner ist, werden n und (n + 1) vertauscht.
// Zwei aufeinander folgende Elemente knnen daher immer noch den gleichen Helligkeitswert haben
// Immer dann, wenn ein Vertauschen stattgefunden hat, muss die while-Schleife von vorne gestartet werden,
// da noch ein spter folgendes Element kleiner als der Vorgnger sein kann.
// Erst wenn die gesamte Datenreihe erfolgreich den Vergleich bestanden hat (dann bleibt ready = TRUE)
// ist der Elementvergleich erfolgreich beendet

while (ready == FALSE)
	{
	ready = TRUE;

	for (i = 0; i < (cnt - 1); i++)
		{
		j = i + 1;                          // j = i + 1
		if (led_value[i] > led_value[j])		// wenn i > (i + 1)
			{											// dann vertauschen
			t = led_value[i];						// Helligkeitswerte
			led_value[i] = led_value[j];
			led_value[j] = t;

			t = led_bitcode[i];					// zugehrige Portbits
			led_bitcode[i] = led_bitcode[j];
			led_bitcode[j] = t;

			ready = FALSE;							// Abbruchbedingung, wenn eine Umsortierung erforderlich war
			}
		}
	}

// Nun zeitgleich stattfindende Schaltvorgnge finden,
// die BitCodes fr gleichzeitigen Schaltvorgnge zusammenfhren,
// die nachfolgenden Daten vorrcken lassen,
// nach hinten mit 255 auffllen
// den Zhler cnt fr die Gesamtzahl der Eintrge decrementieren.
//
// Zu beachten ist, dass es nicht ausreicht, die die vorgegebenen Helligkeitswerte
// auf Gleichheit zu testen (dann sind die Schaltzeitpunkt definitv identisch).
// Bedingt durch die bersetzungstabelle in OCR1A-Werte und deren sehr langsamen Anstieg
// am Anfang der Tabelle, knnen trotz unterschiedlicher Helligkeitswerte dennoch
// gleich OCR1A-Werte entstehen.
// Auch hier mssen dann die Werte bei Gleichheit zusammengelegt werden.

i = 0;
while (i < (cnt - 1))								// jedes Element mit seinem Nachfolger vergleichen
	{
// Beim Vergleich der Schaltzeitpunkte (OCRA-Werte) reicht es nicht aus, die bermittelten 8-Bit
// Helligkeitswerte zu vergleichen. Gerade am unteren Tabellenende sind die OCRA-Werte mehrerer
// Helligkeitswerte gleich.
// Daher wird gleich ein direkter Vergleich der 16-Bit-OCRA-Werte durchgefhrt

	if (pgm_read_word(&ocra_tab[ led_value[i] ]) == pgm_read_word(&ocra_tab[ led_value[i + 1] ]))
		{
		led_bitcode[i] += led_bitcode[i+1]; // bitcodes addieren = die Portpins zusammenfhren

		for (j = i + 1; j < cnt - 1; j++)   // alle Nachfolgenden zeitlich nach vorne aufrcken lassen
			{
			led_value[j]   = led_value[j + 1];
			led_bitcode[j] = led_bitcode[j + 1];
			}
		cnt--;                     			// jetzt haben wir ein Element weniger
		led_value[cnt] = 255;					// hinten eine '255' auffllen
		led_bitcode[cnt] = 0;					// hinten eine '0' auffllen (der Wert ist ohne Bedeutung)
		}

	else i++;										// wenn keine Gleichheit, dann zum nchsten Element
	}

// Um die Beschickung des Timers1 ohne weitere Rechenoperationen komplett ber die Arrays steuern zu knnen,
// muss ein Element an der Startposition vorhanden sein.
// Wenn alle LED's eingeschaltet sind, dann enthlt die Bitmaske fr die zu schaltenden LEDS
// genau ein Bit fr jede LED (damit diese am PWM-Start eingeschaltet wird).
// Bei LED's, die whrend des gesamten PWM-Taktes abgeschaltet sind, muss jeweils ein 0-Bit eingetragen sein.
// Wenn keine der LEDs ausgeschaltet ist, dann muss ein Array-Element fr den PWM-Start angelegt werden.
// Nur led_value[0] mit der geringsten Helligkeit nach der Sortierung und Zusammenfassung kann eine 0 enthalten,
// daher ist es ausreichend diesen Wert zu prfen
if (led_value[0] > 0)      					// wenn das erste Element nicht 0 ist
	{                    						// dann einen Wert fr 0 einfgen
	i = (++cnt);           						// dann gibt es 1 Element mehr im Array
	while (i )            						// dazu alles um eine Stelle nach rechts schieben
		{
		led_value[i]   = led_value[i - 1];
		led_bitcode[i] = led_bitcode[i- 1];
		i--;
		}
	led_value[0]   = 0;       					// hier wird der Wert fr den Start eingefgt
	led_bitcode[0] = LED_PINS; 				// alle LEDs an, daher alle Pins setzen
	}

else													// mindestens eine LED hat den Helligkeitswert 0
	{                                      // erst einmal alle LED_Pins setzten, aber davon
	led_bitcode[0] = LED_PINS - led_bitcode[0]; // die 0-Bits abziehen.
	}

// Im Normalfall sind beim PWM-Start die Port-Bits aller LED's gesetzt.
// Immmer wenn eine LED abgeschaltet werden muss, wird das zugehrige Bit gelscht.
// Im Normalfall wrde man den Portstatus einlesen, mit einer Bitmaske undieren und wieder
// ins Register des Port zurckschreiben. Das dauert aber zu lange.
// Darum wird fr jeden Schaltvorgang die bentigte Bitkonstellation vorausberechnet und
// im Array pwm_buffer[] abgelegt.
// Wird nun der OCR1A-Interrupt aufgerufen, dann muss die ISR nur noch ein Byte in das Portregister
// schreiben - das spart einige CPU-Takte.
// Im Array bitcode[] entsteht dieses Statusbyte.
// Bei Start des PWM-Taktes wird fr alle nicht abgeschalteten LED's eine 1 gesetzt.
// Immer dann, wenn eine LED abgeschaltet werden muss, wird einfach ihr bit_code[n] vom
// Vorgnger bitcode[n-1] subtrahiert.

for (i = 1; i < cnt; i++)
	{
	led_bitcode[i] = ( led_bitcode[i-1] - led_bitcode[i] );// in diesem Fall kann auch subtrahiert werden
	}
														// hier werden die OCR1A-Werte und die Portbits seriell in
														// der spter bentigten Reihenfolge in ein Array geschrieben.
														// Das Array ist gedanklich in zwei Hlften/Speicherbnke geteilt:
														// whrend in die untere Hlfte Daten geschrieben werden knnen,
														// kann die ISR aus der anderen Hlfte die Werte auslesen.
														// Nur der Index entscheidet, wer welche Hlfte benutzt.
if (GPIOR0 == 0) i = 32;		            // wenn der Index < 32, dann die oberen Variablen benutzten
else i = 0;

j = k = 0;											// Die ISR nutzen als Zhlervariable das Register GPIOR0.
while(k <= (LEDs))
	{
	tt = pgm_read_word(&ocra_tab[ led_value[k + 1] ]); // den OCR1A-Wert der Phase (n+1) schreiben

	pwm_buffer[i + (j++)] = (uint8_t) (tt >> 8);
	pwm_buffer[i + (j++)] = (uint8_t) (tt);
#ifdef HIGHSIDE                                 // den Bitcode der Phase (n) schreiben
	pwm_buffer[i + (j++)] = ~led_bitcode[k++] ;	// wenn zum EINgeschalten der Pin auf low gezogen werden muss
#else
	pwm_buffer[i + (j++)] =  led_bitcode[k++] ;	// wenn zum EINgeschalten der Pin auf high gezogen werden muss
#endif
	}


pwm_update = TRUE;                        // damit wird die Speicherbank in der ISR gewechselt
                                          // die ISR ndert den Wert in GPIOR0 [1/0]

while(pwm_update);                        // wartet, bis zum Beginn des nchsten PWM-Taktes
                                          // die neuen Werte sind dann bereits zugewiesen und aktiv
                                          // und die andere Speicherbank kann beschrieben werden

}

// ----------------------------------------------------------------------------
// Farbrampe zwischen den drei Grundfarben mittels fade()
void demo1(void)
{
while(1)
   {
   led_value[0]= 255;
   led_value[1]= 0;
   led_value[2]= 0;
   fade(255);
   if (rx_flag) break;

   led_value[0]= 0;
   led_value[1]= 255;
   led_value[2]= 0;
   fade(255);
   if (rx_flag) break;

   led_value[0]= 0;
   led_value[1]= 0;
   led_value[2]= 255;
   fade(255);
   if (rx_flag) break;

   led_value[0]= 200;
   led_value[1]= 200;
   led_value[2]= 200;
   fade(255);
   if (rx_flag) break;
   }
led_value[0]= 200;
led_value[1]= 200;
led_value[2]= 200;
fade(255);
}

// ----------------------------------------------------------------------------
// Testroutine, hier wird eigentlich nur geblinkt bzw ein schneller bergang ausgefhrt
void demo2(void)
{
uint8_t t, x, y, z;

x = 255;
y = 0;
z = 0;

while(rx_flag == FALSE)
   {
   t = x;
   x = y;
   y = z;
   z = t;
   led_value[0] = x;
   led_value[1] = y;
   led_value[2] = z;

   fade(15);                  // schneller bergang
//   set_light();             // das fhrt zum Blinken
//   _delay_ms(30);
	}

led_value[0] = 200;
led_value[1] = 200;
led_value[2] = 200;
set_light();

}

// ----------------------------------------------------------------------------
// berblenden zwischen 6 Grundfarben mittels fade()

void demo3(void)
{
uint8_t x = 0;

while(rx_flag == FALSE)
   {
   if (x == 0)
      {
      x  = 1;
      led_value[0]= 255;
      led_value[1]= 0;
      led_value[2]= 0;
      }
   else if (x == 1)
      {
      x  = 2;
      led_value[0]= 235;
      led_value[1]= 235;
      led_value[2]= 0;
      }

   else if (x == 2)
      {
      x  = 3;
      led_value[0]= 235;
      led_value[1]= 235;
      led_value[2]= 0;
      }

   else if (x == 3)
      {
      x  = 4;
      led_value[0]= 0;
      led_value[1]= 255;
      led_value[2]= 0;
      }

   else if (x == 4)
      {
      x  = 5;
      led_value[0]= 0;
      led_value[1]= 235;
      led_value[2]= 235;
      }

   else if (x == 5)
      {
      x  = 6;
      led_value[0]= 0;
      led_value[1]= 235;
      led_value[2]= 235;
      }

   else if (x == 6)
      {
      x  = 7;
      led_value[0]= 0;
      led_value[1]= 0;
      led_value[2]= 255;
      }

   else if (x == 7)
      {
      x  = 8;
      led_value[0]= 235;
      led_value[1]= 0;
      led_value[2]= 235;
      }

   else if (x == 8)
      {
      x  = 0;
      led_value[0]= 235;
      led_value[1]= 0;
      led_value[2]= 235;
      }
   fade(75);
   }

led_value[0]= 200;
led_value[1]= 200;
led_value[2]= 200;
fade(75);

}

// ----------------------------------------------------------------------------
// Testroutine fr 3 versetzte/gegenlufige Farbrampen
// mittels direkter Ausgabe set_light()
void demo4(void)
{
uint8_t x, y, z;

x = 20;
y = 128;
z = 150;

while(rx_flag == FALSE)
   {
   led_value[0] = x++;
   led_value[1] = y++;
   led_value[2] = z++;
   set_light();
	}

led_value[0] = 200;
led_value[1] = 200;
led_value[2] = 200;
set_light();
}
// --- [ eof ] ----------------------------------------------------------------
